
#include <windows.h>
#include <windowsx.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "dibtools.h"
#include "win2tif.h"

/* define the supported format */
/* By deflaut : Jpeg, Png, Bmp, Emf, Wmf and Gif */



#if !defined(__BMP__) && !defined(__NOBMP__)
#define __BMP__
#endif


/* step for expanding memory */
/* when the Bitmap in construction has filled all the buffer, we expand
   the buffer by STEPEXPAND* bytes */
#define STEPEPANDSTANDARD   (0x8000L)
#define STEPEXPANDPNG       (STEPEPANDSTANDARD)
#define STEPEXPANDTIF       (STEPEPANDSTANDARD)
#define STEPEXPANDGIF       (STEPEPANDSTANDARD)
#define STEPEXPANDJPEG      (STEPEPANDSTANDARD)
#define STEPEXPANDPCX       (STEPEPANDSTANDARD)

/*****************************************************/

#ifdef __LIBTIFF__
#include "tiffio.h"
#include "tiffiop.h"
#endif

/*****************************************************/

/*****************************************************/

#ifdef __BMP__
DWORD FlushDibInBmp(LPVOID lpDib,LPVOID lpRawData,long lQuality,
                    LPVOID lpPtr,DWORD dwSize,
                    TResizeFlushPtrFunc pResizeFlushPtrFunc,
                    LPVOID lpCustomDataForResize,LPVOID* lplpNewPtr,
                    BOOL fDibInMemory = FALSE);

#ifndef NOPAINTBITMAP
class PAINTTOFORMATBMP:public PAINTBITMAP
{
public:
    PAINTTOFORMATBMP();
    ~PAINTTOFORMATBMP() ;
    BOOL Init(long x,long y,long lNbBitsColor,
                const COLORREF* pcrPalette,long iNbColorPalette,long lQuality);
    HDC  GetPaintDC();
    DWORD FlushTif(LPVOID lpPtr,DWORD dwSize,
                   TResizeFlushPtrFunc pResizeFlushPtrFunc,
                   LPVOID lpCustomDataForResize,LPVOID* lplpNewPtr);
	const char* GetMimeType() { return "image/bmp" ; } ;
    void  SetDibInMemory() ;
private:
    BYTE	szHeader[sizeof(BITMAPINFO)+((256+16)*sizeof(RGBQUAD))];
    long    lX,lY,lNbBitsColor;
    HDC     hDC;
    HBITMAP hbm;
    unsigned char *rawdata;
	long	iNbColorPalette;
    BOOL    fDibInMemory;
};
#endif
#endif
/*****************************************************/

#ifdef __LIBTIFF__
DWORD FlushDibInTiff(LPVOID lpDib,LPVOID lpRawData,long lQuality,
                    LPVOID lpPtr,DWORD dwSize,
                    TResizeFlushPtrFunc pResizeFlushPtrFunc,
                    LPVOID lpCustomDataForResize,LPVOID* lplpNewPtr);

#ifndef NOPAINTBITMAP
class PAINTTOFORMATTIFF:public PAINTBITMAP
{
public:
    PAINTTOFORMATTIFF();
    ~PAINTTOFORMATTIFF() ;
    BOOL Init(long x,long y,long lNbBitsColor,
                const COLORREF* pcrPalette,long iNbColorPalette,long lQuality);
    HDC  GetPaintDC();
    DWORD FlushTif(LPVOID lpPtr,DWORD dwSize,
                   TResizeFlushPtrFunc pResizeFlushPtrFunc,
                   LPVOID lpCustomDataForResize,LPVOID* lplpNewPtr);
	const char* GetMimeType() { return "image/tiff" ; } ;
private:
    BYTE	szHeader[sizeof(BITMAPINFO)+(256*sizeof(RGBQUAD))];
    long    lX,lY,lNbBitsColor;
    HDC     hDC;
    HBITMAP hbm;
    unsigned char *rawdata;
	long	iNbColorPalette;
	long	lQuality;
};
#endif
#endif



/*****************************************************/

extern "C"
{

#ifndef NOPAINTBITMAP


FLUSHDIBAPIEXPORT PAINTBITMAP* FLUSHDIBAPI GetPaintBitmap(TYPEFORMATOUT tFormatOut)
{
	PAINTBITMAP* pBmp=NULL;
	switch (tFormatOut)
	{
        #ifdef __LIBTIFF__
		case FORMAT_BITMAP_TIFF :
			pBmp=new PAINTTOFORMATTIFF;
			break;
        #endif

        #ifdef __BMP__
		case FORMAT_BITMAP_BMP :
			pBmp=new PAINTTOFORMATBMP;
			break;

        case FORMAT_BITMAP_DIBINMEMORY:
            {
			    PAINTTOFORMATBMP *pBmpDib=new PAINTTOFORMATBMP;
                if (pBmpDib!=NULL)
                    pBmpDib -> SetDibInMemory();
                pBmp = pBmpDib;
            }
			break;
        #endif

	}
	return pBmp;
}


FLUSHDIBAPIEXPORT BOOL FLUSHDIBAPI InitPaintBitmap(LPPAINTBITMAP pBmp, long x,long y,long lNbBitsColor,
                const COLORREF* pcrPalette,long iNbColorPalette,
				long lQuality/*=0*/)
{
    BOOL fRet = FALSE;
    if (pBmp!=NULL)
        fRet = pBmp->Init(x,y,lNbBitsColor,pcrPalette,iNbColorPalette,lQuality);
    return fRet;
}

FLUSHDIBAPIEXPORT HDC  FLUSHDIBAPI GetPaintDCPaintBitmap(LPPAINTBITMAP pBmp)
{
    HDC hDCRet = NULL;
    if (pBmp!=NULL)
        hDCRet = pBmp->GetPaintDC();
    return hDCRet ;
}

FLUSHDIBAPIEXPORT DWORD FLUSHDIBAPI FlushPaintBitmap(LPPAINTBITMAP pBmp, LPVOID lpPtr,DWORD dwSize,
                            TResizeFlushPtrFunc pResizeFlushPtrFunc/*=NULL*/,
                            LPVOID lpCustomDataForResize/*=NULL*/,
                            LPVOID* lplpNewPtr/*=NULL*/)
{
    DWORD dwRet=0;
    if (pBmp!=NULL)
        dwRet = pBmp->FlushTif(lpPtr,dwSize,pResizeFlushPtrFunc,
                            lpCustomDataForResize,lplpNewPtr);
    return dwRet;
}

FLUSHDIBAPIEXPORT const char* FLUSHDIBAPI GetMimeTypePaintBitmap(LPPAINTBITMAP pBmp)
{
    const char* lpRet=NULL;
    if (pBmp!=NULL)
        lpRet = pBmp->GetMimeType();
    return lpRet;
}

FLUSHDIBAPIEXPORT BOOL FLUSHDIBAPI DeletePaintBitmap(LPPAINTBITMAP pBmp) 
{
    BOOL fRet = (pBmp!=NULL);
    if (pBmp!=NULL)
        delete pBmp;
    return fRet;
}

#endif // NOPAINTBITMAP

FLUSHDIBAPIEXPORT TYPEFORMATOUT FLUSHDIBAPI GetFormatOutFromExt(LPCSTR szExt)
{
    TYPEFORMATOUT tOut = FORMAT_BITMAP_INVALID;

    if (!lstrcmpi(szExt, ".GIF"))
        tOut = FORMAT_BITMAP_GIF256;        

    if (!lstrcmpi(szExt, ".PCX"))
        tOut = FORMAT_BITMAP_PCX;

    if ((!lstrcmpi(szExt, ".TIF")) || (!lstrcmpi(szExt, ".TIFF")))
        tOut = FORMAT_BITMAP_TIFF;

    if ((!lstrcmpi(szExt, ".BMP")) || (!lstrcmpi(szExt, ".DIB")))
        tOut = FORMAT_BITMAP_BMP;

    if ((!lstrcmpi(szExt, ".JPG")) || (!lstrcmpi(szExt, ".JPEG")))
        tOut = FORMAT_BITMAP_JPEG;

    if (!lstrcmpi(szExt, ".PNG"))
        tOut = FORMAT_BITMAP_PNG;

    if (!lstrcmpi(szExt, ".EMF"))
        tOut = FORMAT_BITMAP_EMF;

    if (!lstrcmpi(szExt, ".WMF"))
        tOut = FORMAT_BITMAP_WMF;

    return tOut;
}

FLUSHDIBAPIEXPORT DWORD FLUSHDIBAPI FlushDibInBitmap(LPVOID lpDib,
                        LPVOID lpRawData,
                        TYPEFORMATOUT tFormatOut, long lQuality,
                        LPVOID lpPtr,DWORD dwSize,
                        TResizeFlushPtrFunc pResizeFlushPtrFunc,
                        LPVOID lpCustomDataForResize,LPVOID* lplpNewPtr)
{
    DWORD dwRet = 0;
    switch (tFormatOut)
    {
        #ifdef __GIF__
		case FORMAT_BITMAP_GIF256 :
			dwRet = FlushDibInGif(lpDib,lpRawData,lQuality,
                                    lpPtr,dwSize,pResizeFlushPtrFunc,
                                    lpCustomDataForResize,lplpNewPtr);
			break;
        #endif

        #ifdef __LIBTIFF__
		case FORMAT_BITMAP_TIFF:
			dwRet = FlushDibInTiff(lpDib,lpRawData,lQuality,
                                    lpPtr,dwSize,pResizeFlushPtrFunc,
                                    lpCustomDataForResize,lplpNewPtr);
			break;
        #endif

        #ifdef __BMP__
		case FORMAT_BITMAP_BMP :
			dwRet = FlushDibInBmp(lpDib,lpRawData,lQuality,
                                    lpPtr,dwSize,pResizeFlushPtrFunc,
                                    lpCustomDataForResize,lplpNewPtr,
                                    FALSE);
			break;

        case FORMAT_BITMAP_DIBINMEMORY:
			dwRet = FlushDibInBmp(lpDib,lpRawData,lQuality,
                                    lpPtr,dwSize,pResizeFlushPtrFunc,
                                    lpCustomDataForResize,lplpNewPtr,
                                    TRUE);
			break;

        #endif

        #ifdef __JPEG__
		case FORMAT_BITMAP_JPEG :
			dwRet = FlushDibInJpeg(lpDib,lpRawData,lQuality,
                                    lpPtr,dwSize,pResizeFlushPtrFunc,
                                    lpCustomDataForResize,lplpNewPtr);
			break;
        #endif // __JPEG__

        #ifdef __PNG__
		case FORMAT_BITMAP_PNG :
			dwRet = FlushDibInPng(lpDib,lpRawData,lQuality,
                                    lpPtr,dwSize,pResizeFlushPtrFunc,
                                    lpCustomDataForResize,lplpNewPtr);
			break;
        #endif

        #ifdef __PCX__
		case FORMAT_BITMAP_PCX :
			dwRet = FlushDibInPcx(lpDib,lpRawData,lQuality,
                                    lpPtr,dwSize,pResizeFlushPtrFunc,
                                    lpCustomDataForResize,lplpNewPtr);
			break;
        #endif
    }
    return dwRet;
}

} // of extern "C"
/****************************************************************************/
/*  SOMES Windows Bitmap tools */


long CreateWebPaletteClrRef(COLORREF* lpColorRef)
{
    long iNbColorPalette;
    iNbColorPalette=6*6*6;
    
    for (BYTE bR=0;bR<6;bR++)
        for (BYTE bG=0;bG<6;bG++)
            for (BYTE bB=0;bB<6;bB++)
    {
        *lpColorRef = RGB(bR*51,bG*51,bB*51);
        lpColorRef++;
    }
    return iNbColorPalette;
}


BOOL InitBitmapForDrawingInside(long lX,long lY,long lNbBitsColor,
                        const COLORREF* pcrPalette,long iNbColorPalette,
						LPBYTE szHeader,DWORD dwSizeHeader,
						HDC& hDC,HBITMAP& hbm,LPBYTE &rawdata)
{
    BITMAPINFO bi;

    memset(szHeader,0x00,(dwSizeHeader));


    // for getpixel


    if ((iNbColorPalette!=0) && (pcrPalette!=NULL))
    {
        memcpy((((BYTE*)szHeader)+sizeof(BITMAPINFOHEADER)),
                    pcrPalette,
                    sizeof(COLORREF)*(iNbColorPalette));
    }
    else if (lNbBitsColor==8)
    {
        COLORREF* lpColorRef = (COLORREF*) (((BYTE*)szHeader)+sizeof(BITMAPINFOHEADER));
        iNbColorPalette=CreateWebPaletteClrRef(lpColorRef);        
    }


    bi.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
    bi.bmiHeader.biWidth = lX;
    bi.bmiHeader.biHeight = lY;
    bi.bmiHeader.biPlanes=1;
    //bi.bmiHeader.biBitCount=(short)lNbColorInit;
    bi.bmiHeader.biBitCount=(WORD)lNbBitsColor; // WORD needed by Windows...

    bi.bmiHeader.biCompression=BI_RGB;
    bi.bmiHeader.biSizeImage=0;
    bi.bmiHeader.biXPelsPerMeter=96;
    bi.bmiHeader.biYPelsPerMeter=96;
    //bi.bmiHeader.biClrUsed=1<<lNbBitsColorInit;
    //bi.bmiHeader.biClrImportant=1<<lNbBitsColorInit;

	//if (iNbColorPalette<=2)
    if (iNbColorPalette<2)
	{
    bi.bmiHeader.biClrUsed=0;
    bi.bmiHeader.biClrImportant=0;
	}
	else
	{
    bi.bmiHeader.biClrUsed=iNbColorPalette;
    bi.bmiHeader.biClrImportant=iNbColorPalette;
	}

    memcpy(szHeader,&bi,sizeof(BITMAPINFOHEADER));

    rawdata = NULL;
    hDC = NULL;
    hbm = NULL;

    hDC= CreateCompatibleDC(NULL);
    if (hDC != NULL)
    {
        hbm=CreateDIBSection(hDC,(BITMAPINFO*)szHeader,DIB_RGB_COLORS,(LPVOID*)&rawdata,NULL,0);

        if (hbm != NULL)    
            SelectBitmap(hDC,hbm);
        else
            DeleteDC(hDC);
    }

    return (hbm != NULL);
}

/****************************************************************************/

static DWORD EndFlushByResizeBuffer(DWORD dwFinalSize, LPVOID lpPtr,
                                    TResizeFlushPtrFunc pResizeFlushPtrFunc,
                                    LPVOID lpCustomDataForResize,
                                    LPVOID* lplpNewPtr)
{
    LPVOID lpSavePtr = lpPtr;

    if (pResizeFlushPtrFunc!=NULL)
    {
        LPVOID lpNewPtr = (*pResizeFlushPtrFunc)(lpPtr,
                                       lpCustomDataForResize,dwFinalSize);
        if ((lpNewPtr!=NULL) || (dwFinalSize==0))
            lpSavePtr=lpNewPtr;
    }
    if (lplpNewPtr!=NULL)
        *lplpNewPtr=lpSavePtr;
    return dwFinalSize;    
}
//#define DIBTool_IS_WIN30_DIB(lpbi)  ((*(LPDWORD)(lpbi)) == sizeof(BITMAPINFOHEADER))


/****************************************************************************/
/****************************************************************************/

#ifdef __BMP__
#ifndef NOPAINTBITMAP
PAINTTOFORMATBMP::PAINTTOFORMATBMP()
{
    hDC = NULL;
    hbm = NULL;

    rawdata=NULL;
    fDibInMemory = FALSE;
}

void PAINTTOFORMATBMP::SetDibInMemory()
{
    fDibInMemory = TRUE;
}

PAINTTOFORMATBMP::~PAINTTOFORMATBMP() 
{
}

BOOL PAINTTOFORMATBMP::Init(long lXInit,long lYInit,long lNbBitsColorInit,
                        const COLORREF* pcrPalette,long iNbColorPaletteInit,
						long)
{
    lX = lXInit;
    lY = lYInit;
    lNbBitsColor = lNbBitsColorInit;
	iNbColorPalette = iNbColorPaletteInit;
    memset(szHeader,0,sizeof(szHeader));

	// For BW, we don't pack in 1 bit/pixel (I don't know why, but we obtain full black image)!
	InitBitmapForDrawingInside(lX,lY,
						lNbBitsColor,//(lNbBitsColor==1) ? 4 : lNbBitsColor,
                        pcrPalette,
						iNbColorPalette,
						szHeader,sizeof(szHeader),
						hDC,hbm,rawdata);



    return (hbm != NULL);
}

HDC  PAINTTOFORMATBMP::GetPaintDC()
{
    return hDC;
}

DWORD PAINTTOFORMATBMP::FlushTif(LPVOID lpPtr,DWORD dwSize,
                                 TResizeFlushPtrFunc pResizeFlushPtrFunc,
                                 LPVOID lpCustomDataForResize,LPVOID* lplpNewPtr)
{
    DWORD dwRet=0;
    GdiFlush();
	if (hbm!=NULL)
	{
        dwRet=FlushDibInBmp(szHeader,rawdata,0,
                      lpPtr,dwSize,
                      pResizeFlushPtrFunc,
                      lpCustomDataForResize,lplpNewPtr,
                      fDibInMemory);
	}


    DeleteDC(hDC);
    DeleteBitmap(hbm);

    return dwRet;
}
#endif // NOPAINTBITMAP



DWORD FlushDibInBmp(LPVOID lpDib,LPVOID lpRawData,long lQuality,
                    LPVOID lpPtr,DWORD dwSize,
                    TResizeFlushPtrFunc pResizeFlushPtrFunc,
                    LPVOID lpCustomDataForResize,LPVOID* lplpNewPtr,
                    BOOL fDibInMemory)
{
    long    lX,lY,lNbBitsColor;
    LPBYTE  rawdata,rawdata_afterdib;
	long	iNbColorPalette;
    unsigned char *szHeader;

    DIBTool_GetDIBInfo(lpDib,lX,lY,lNbBitsColor,iNbColorPalette,rawdata_afterdib);
    szHeader = (LPBYTE)lpDib;
    rawdata = (lpRawData == NULL) ? rawdata_afterdib : ((LPBYTE)lpRawData);

    DWORD dwSizeAllFile=0;

	{
		DWORD dwSizeHeader;
		DWORD dwSizeBits;

        dwSizeHeader=Round4(sizeof(BITMAPINFOHEADER)+(sizeof(RGBQUAD)*iNbColorPalette));
        dwSizeBits = Round4(((lX*lNbBitsColor)+7)/8)*lY;
        dwSizeAllFile = sizeof(BITMAPFILEHEADER)+dwSizeHeader+dwSizeBits;

        if (dwSizeAllFile>dwSize)
        {
            LPVOID lpNewPtr =NULL;
            if (pResizeFlushPtrFunc != NULL)
                lpNewPtr = (*pResizeFlushPtrFunc)(lpPtr,lpCustomDataForResize,dwSizeAllFile);
            if (lpNewPtr != NULL)
                lpPtr = lpNewPtr ;
            else
                dwSizeAllFile = 0;
        }

        if (dwSizeAllFile > 0)
        {
		    BITMAPFILEHEADER* lpBmfh=(BITMAPFILEHEADER*)lpPtr;
		    LPBITMAPINFO lpBmi;

            if (fDibInMemory)
                lpBmi=(LPBITMAPINFO)(lpBmfh);
            else
            {
                lpBmi=(LPBITMAPINFO)(lpBmfh+1);
		        lpBmfh->bfType=0x4d42;
		        
		        lpBmfh->bfReserved1=0;
		        lpBmfh->bfReserved2=0;
		        lpBmfh->bfOffBits=sizeof(BITMAPFILEHEADER)+dwSizeHeader;
		        lpBmfh->bfSize=dwSizeAllFile;
            }
		    memcpy(lpBmi,szHeader,dwSizeHeader);
		    lpBmi->bmiHeader.biBitCount=(WORD)lNbBitsColor; // WORD needed by Windows...

		    //if (lNbBitsColor>1)
			    memcpy(((LPBYTE)lpBmi)+dwSizeHeader,rawdata,dwSizeBits);
            /*
		    else
		    {
			    for (long i=0;i<lY;i++)
			    {
				    LPBYTE lpBaseReadLine = rawdata + (i*Round4((lX+1)/2));
				    LPBYTE lpBaseWriteLine = (((LPBYTE)lpBmi)+dwSizeHeader) + (Round4((lX+7)/8)*i);
				    long lNbDoingRead=(lX+7)/8;
				    for (long j=0;j<lNbDoingRead;j++)
				    {
					    *lpBaseWriteLine =
						    ((((*(lpBaseReadLine+0)) & 0xf0) != 0) ? 0x80 : 0) |
						    ((((*(lpBaseReadLine+0)) & 0x0f) != 0) ? 0x40 : 0) |
						    ((((*(lpBaseReadLine+1)) & 0xf0) != 0) ? 0x20 : 0) |
						    ((((*(lpBaseReadLine+1)) & 0x0f) != 0) ? 0x10 : 0) |
						    ((((*(lpBaseReadLine+2)) & 0xf0) != 0) ? 0x08 : 0) |
						    ((((*(lpBaseReadLine+2)) & 0x0f) != 0) ? 0x04 : 0) |
						    ((((*(lpBaseReadLine+3)) & 0xf0) != 0) ? 0x02 : 0) |
						    ((((*(lpBaseReadLine+3)) & 0x0f) != 0) ? 0x01 : 0) ;
					    lpBaseReadLine+=4;
					    lpBaseWriteLine++;
				    }
				    *lpBaseWriteLine = *(lpBaseWriteLine+1) = *(lpBaseWriteLine+2) = 0; // clean-up (round4!)
			    }
		    }*/
        }
	}


    if ((lplpNewPtr!=NULL))
        *lplpNewPtr = lpPtr;

    return dwSizeAllFile;
}
#endif // __BMP__

/****************************************************************************/



/****************************************************************************/

// PAINTTOFORMATTIFF
#ifdef __LIBTIFF__
#ifndef NOPAINTBITMAP
PAINTTOFORMATTIFF::PAINTTOFORMATTIFF()
{
    hDC = NULL;
    hbm = NULL;

    rawdata=NULL;
}

PAINTTOFORMATTIFF::~PAINTTOFORMATTIFF() 
{
}

BOOL PAINTTOFORMATTIFF::Init(long lXInit,long lYInit,long lNbBitsColorInit,
                        const COLORREF* pcrPalette,long iNbColorPaletteInit,
						long lQualitySet)
{
    lX = lXInit;
    lY = max(1,lYInit); // LibPNG don't like 0 size in Y!
    lNbBitsColor = lNbBitsColorInit;
	iNbColorPalette = iNbColorPaletteInit;
	lQuality = lQualitySet;

    InitBitmapForDrawingInside(lX,lY,(lNbBitsColor==1) ? 1:lNbBitsColor,
                        pcrPalette,iNbColorPalette,
						szHeader,sizeof(szHeader),
						hDC,hbm,rawdata);



    return (hbm != NULL);
}

HDC  PAINTTOFORMATTIFF::GetPaintDC()
{
    return hDC;
}

DWORD PAINTTOFORMATTIFF::FlushTif(LPVOID lpPtr,DWORD dwSize,
                                 TResizeFlushPtrFunc pResizeFlushPtrFunc,
                                 LPVOID lpCustomDataForResize,LPVOID* lplpNewPtr)
{
    DWORD dwRet=0;
    GdiFlush();
	if (hbm!=NULL)
	{
        dwRet=FlushDibInTiff(szHeader,rawdata,lQuality,
                      lpPtr,dwSize,
                      pResizeFlushPtrFunc,
                      lpCustomDataForResize,lplpNewPtr);
	}


    DeleteDC(hDC);
    DeleteBitmap(hbm);

    return dwRet;
}
#endif // NOPAINTBITMAP


typedef struct
{
  unsigned char * pData;
  DWORD    FileSize;
  DWORD    CurPos;
  DWORD    LastWritten;
  TResizeFlushPtrFunc pResizeFlushPtrFunc;
  LPVOID lpCustomDataForResize;
  LPVOID* lplpNewPtr;
  LPDWORD lpdwSizeFinal;
} MemDestTIFFInfo;


static tsize_t
_tiffReadProc(thandle_t fd, tdata_t buf, tsize_t size)
{
	MemDestTIFFInfo * pInfo = (MemDestTIFFInfo *) fd;

  // Make sure we don't run over the end of the file.
  if (size+pInfo->CurPos > pInfo->FileSize)
    size = pInfo->FileSize-pInfo->CurPos;

  memcpy ((void *)buf, (void *)(pInfo->pData+pInfo->CurPos), size);
  pInfo->CurPos += size;

	return (tsize_t) size;
}


static tsize_t
_tiffWriteProc(thandle_t fd, tdata_t buf, tsize_t length)
{
    //WRITEINFO *pWriteInfo=(WRITEINFO *)png_get_io_ptr(png_ptr);
    MemDestTIFFInfo * pInfo = (MemDestTIFFInfo *) fd;
    if ((length + pInfo->CurPos > pInfo->FileSize) && 
        (pInfo->pResizeFlushPtrFunc != NULL))
    {
        LPVOID lpNewPtr ;
        DWORD dwExpand = MyRound(((length + pInfo->CurPos) - pInfo->FileSize),
                            STEPEXPANDTIF);
        lpNewPtr = (*(pInfo->pResizeFlushPtrFunc))(pInfo->pData,
                                          pInfo->lpCustomDataForResize,
                                          pInfo->FileSize + dwExpand);
        if (lpNewPtr != NULL)
        {
            pInfo->pData = (unsigned char*)lpNewPtr ;
            pInfo->FileSize += dwExpand;
        }
    }

    length=min(length,(tsize_t)(pInfo->FileSize - pInfo->CurPos));
    memcpy(pInfo->pData+pInfo->CurPos,buf,length);
    pInfo->CurPos+=length;
    pInfo->LastWritten=max(pInfo->LastWritten,pInfo->CurPos);
    return length;
}

static toff_t
_tiffSeekProc(thandle_t fd, toff_t off, int whence)
{
  MemDestTIFFInfo * pInfo = (MemDestTIFFInfo *) fd;

	switch(whence)
	  {
	    case 1:
		    pInfo->CurPos += off;
		    break;
	    case 2:
        pInfo->CurPos = pInfo->LastWritten-off;
		    break;
	    case 0:
	    default:
        pInfo->CurPos = off;
		    break;
	  }

if (pInfo->CurPos>pInfo->LastWritten)
{
    OutputDebugString("err\n");
}
	return pInfo->CurPos;
}

static int
_tiffCloseProc(thandle_t fd)
{
  MemDestTIFFInfo * pInfo = (MemDestTIFFInfo *) fd;
  if (pInfo->lplpNewPtr!=NULL)
      *(pInfo->lplpNewPtr)=pInfo->pData;
  if (pInfo->lpdwSizeFinal!=NULL)
      *pInfo->lpdwSizeFinal=pInfo->LastWritten;
  free (pInfo);

	return 0;
}

static toff_t
_tiffSizeProc(thandle_t fd)
{
  MemDestTIFFInfo * pInfo = (MemDestTIFFInfo *) fd;
	return ((toff_t)(pInfo->LastWritten));
}

static int
_tiffDummyMapProc(thandle_t fd, tdata_t* pbase, toff_t* psize)
{
    /*
    MemDestTIFFInfo * pInfo = (MemDestTIFFInfo *) fd;
    *pbase = pInfo->pData;
    *psize = pInfo->FileSize;
    return(1);
    */
	return(0);
}

static void
_tiffDummyUnmapProc(thandle_t fd, tdata_t base, toff_t size)
{
	return;
}

TIFF*
TifOpenMemWrite (unsigned char * lpPtr, DWORD FileSize,
             TResizeFlushPtrFunc pResizeFlushPtrFunc,
             LPVOID lpCustomDataForResize,LPVOID* lplpNewPtr,
             LPDWORD lpdwSizeFinal)
{
	TIFF* tif;
  MemDestTIFFInfo * pInfo;

  pInfo = (MemDestTIFFInfo *) malloc (sizeof (MemDestTIFFInfo));
  pInfo->pData = lpPtr;
  pInfo->CurPos = 0;
  pInfo->FileSize = FileSize;
  pInfo->pResizeFlushPtrFunc=pResizeFlushPtrFunc,
  pInfo->lpCustomDataForResize=lpCustomDataForResize;
  pInfo->lplpNewPtr=lplpNewPtr;
  pInfo->lpdwSizeFinal = lpdwSizeFinal;
  pInfo->LastWritten=0;

	tif = TIFFClientOpen ("MemDest", "w", (thandle_t)pInfo,
	                     _tiffReadProc, _tiffWriteProc,
	                     _tiffSeekProc, _tiffCloseProc, _tiffSizeProc,
		                   _tiffDummyMapProc, _tiffDummyUnmapProc);
	if (tif)
		tif->tif_fd = (int)(thandle_t)pInfo;
	return (tif);
}


DWORD FlushDibInTiff(LPVOID lpDib,LPVOID lpRawData,long lQuality,
                    LPVOID lpPtr,DWORD dwSize,
                    TResizeFlushPtrFunc pResizeFlushPtrFunc,
                    LPVOID lpCustomDataForResize,LPVOID* lplpNewPtr)
{
    DWORD dwRet=0;
    long    lX,lY,lNbBitsColor,lXByte;
    LPBYTE  rawdata,rawdata_afterdib;
	long	iNbColorPalette;
    BOOL fIsPalette;
    DWORD dwNbPalEntry;
    BOOL fIsGrayPalette;
    BOOL fIsOrderedGrayPalette;
    DWORD dwStepPalInGray;
    TIFF* tif;
    RGBQUAD* lpPalDib=NULL;
    LPVOID lpDummy=NULL;

    if (lplpNewPtr==NULL)
        lplpNewPtr=&lpDummy;

    DIBTool_GetDIBInfo(lpDib,lX,lY,lNbBitsColor,iNbColorPalette,rawdata_afterdib);
    if (lNbBitsColor<=8) 
        lpPalDib = (((LPBITMAPINFO)lpDib)->bmiColors) ;
    DIBTool_AnalysePalette( lpPalDib,
                            lNbBitsColor,
                            fIsPalette, dwNbPalEntry,
                            fIsGrayPalette, fIsOrderedGrayPalette,
                            dwStepPalInGray);

    rawdata = (lpRawData == NULL) ? rawdata_afterdib : ((LPBYTE)lpRawData);
    //fIsGrayPalette=FALSE;

    tif=TifOpenMemWrite((unsigned char*)lpPtr,dwSize,
                         pResizeFlushPtrFunc,lpCustomDataForResize,
                         lplpNewPtr,&dwRet);
    if (tif!=NULL) 
    {
        BOOL fIsRgb = lNbBitsColor>=24;
        DWORD i;
        DWORD dwCompression ;
        BOOL fInvertBilevel=FALSE;
        
        unsigned char*lpCurLineBuffer =NULL;
        TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, lX);
        TIFFSetField(tif, TIFFTAG_IMAGELENGTH, lY);
        TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, fIsRgb ? 8 : lNbBitsColor);

        dwCompression = COMPRESSION_LZW;
        if ((lQuality>=COMPRESSION_NONE) && (lQuality<=COMPRESSION_JPEG))
            dwCompression = (DWORD)lQuality;
        TIFFSetField(tif, TIFFTAG_COMPRESSION, dwCompression);//NONE);

        if ((lNbBitsColor>=8) && (dwCompression == COMPRESSION_LZW) && (lQuality!=50))
            TIFFSetField(tif, TIFFTAG_PREDICTOR, 2);

        if ((lNbBitsColor==1) && 
            ((dwCompression == COMPRESSION_CCITTFAX3) || 
             (dwCompression == COMPRESSION_CCITTFAX4) ||
             (dwCompression == COMPRESSION_CCITTRLE)))
            fInvertBilevel=TRUE;


        if ((lNbBitsColor==1))
            fInvertBilevel=TRUE;

        // we set predictor to 2 to enhance LZW compression. See page 64 of tiff6.pdf
        //  (the TIFF 6.0 specification June 3, 1992 from Adobe)
        
        TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, fIsRgb ? PHOTOMETRIC_RGB :
                        (fIsOrderedGrayPalette ?
                            (fInvertBilevel ? PHOTOMETRIC_MINISWHITE : PHOTOMETRIC_MINISBLACK) :
                                    PHOTOMETRIC_PALETTE));
        TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, fIsRgb ? 3 : 1);
        TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, 1);
        TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
        TIFFSetField(tif, TIFFTAG_RESOLUTIONUNIT, RESUNIT_NONE);

        if (fIsOrderedGrayPalette)
        {
            //TIFFSetField(tif, TIFFTAG_REFERENCEBLACKWHITE, refblackwhite);
            //TIFFSetField(tif, TIFFTAG_TRANSFERFUNCTION, gray);
        }
        if ((fIsPalette) && (!fIsOrderedGrayPalette) && (lpPalDib!=NULL))
        {
            u_short         *red, *green, *blue;
            u_short         *alloc_pal;
            alloc_pal = (u_short*)malloc(sizeof(u_short)*(3*0x100));
            if (alloc_pal!=NULL)
            {
                red = alloc_pal;
                green = red + 0x100;
                blue = green+0x100;
                DWORD i;
                for (i=0;i<dwNbPalEntry;i++)
                {
                    *(red+i) =  (lpPalDib+i)->rgbRed << 8;
                    *(blue+i) = (lpPalDib+i)->rgbBlue << 8;
                    *(green+i) =(lpPalDib+i)->rgbGreen << 8;
                }

                TIFFSetField(tif, TIFFTAG_COLORMAP, red, green, blue);
                free(alloc_pal);
            }
        }
        if (fIsRgb)
        {
            //TIFFSetField(tif, TIFFTAG_REFERENCEBLACKWHITE, refblackwhite);
            //TIFFSetField(tif, TIFFTAG_TRANSFERFUNCTION, red, green, blue);
        }

       if (lNbBitsColor>8)
           lXByte = Round4(lX*3);
       else if (lNbBitsColor>4)
           lXByte = Round4(lX);
       else if (lNbBitsColor>1)
           lXByte = Round4((lX+1)/2);
       else if (lNbBitsColor==1)
           lXByte = Round4((lX+7)/8);

       if ((lNbBitsColor>8) || fInvertBilevel)
       {
           lpCurLineBuffer = (unsigned char*)malloc((lX*3)+0x20);
       }

        for (i = 0; i < (DWORD)lY; i++)
        {
            unsigned char * scan_line;
            scan_line = rawdata + ( (lY-((i*1)+1))  *lXByte);

            if ((fIsRgb) && (lpCurLineBuffer!=NULL))
            {
                    unsigned char* lpBaseParcLineRead = (unsigned char*)scan_line;
                    unsigned char* lpBaseParcLineWrite = lpCurLineBuffer;
                    long j;
                    for (j=0;j<lX;j++)
                    {
                            *(lpBaseParcLineWrite+0)=*(lpBaseParcLineRead+2);
                            *(lpBaseParcLineWrite+1)=*(lpBaseParcLineRead+1);
                            *(lpBaseParcLineWrite+2)=*(lpBaseParcLineRead+0);
                            lpBaseParcLineRead+=3;
                            lpBaseParcLineWrite+=3;
                    }
                    scan_line = lpCurLineBuffer;
            }
            else if (fInvertBilevel)
            {
                    unsigned char* lpBaseParcLineRead = (unsigned char*)scan_line;
                    unsigned char* lpBaseParcLineWrite = lpCurLineBuffer;
                    long j;
                    for (j=0;j<((lX+7)/8);j++)
                    {
                            *(lpBaseParcLineWrite)=~(*(lpBaseParcLineRead));
                            lpBaseParcLineRead+=1;
                            lpBaseParcLineWrite+=1;
                    }
                    scan_line = lpCurLineBuffer;
            }

            TIFFWriteScanline(tif, scan_line, i, 0);
        }
        if (lpCurLineBuffer!=NULL)
            free(lpCurLineBuffer);
        lpCurLineBuffer=NULL;
        TIFFClose(tif);
    }

    return EndFlushByResizeBuffer(dwRet,*lplpNewPtr,
                        pResizeFlushPtrFunc,lpCustomDataForResize,lplpNewPtr);
   //return dwRet;

}

#endif // LIBTIFF

/****************************************************************************/

/****************************************************************************/
