/* Win2Tif.h : tools for writing Bitmap Files */
/* The TIF in the name is historical */


// FORMAT_BITMAP_WMF and FORMAT_BITMAP_EMF not supported by FlushDibInBitmap, but by GetPaintBitmap

// begin of definition of FLUSHDIBAPI, FLUSHDIBAPIEXPORT and FLUSHDIBCLASS

#define FLUSHDIBAPI WINAPI

#if defined (__XGOBMP__) || defined (GCISINTERNAL) || defined (__FLUSHDIB__)

#if defined( __BORLANDC__ )
#define FLUSHDIBAPIEXPORT _export
#elif  defined(_MSC_VER)
#define FLUSHDIBAPIEXPORT __declspec( dllexport )
#elif
#error NO KNOWN COMPILER
#endif

#else

#if defined( __BORLANDC__ )
#define FLUSHDIBAPIEXPORT _import
#elif  defined(_MSC_VER)
#define FLUSHDIBAPIEXPORT __declspec( dllimport )
#elif
#error NO KNOWN COMPILER
#endif

#endif

#define FLUSHDIBCLASS     FLUSHDIBAPIEXPORT

////  end

typedef enum
{
    FORMAT_BITMAP_GIF256,
    FORMAT_BITMAP_BMP,
    FORMAT_BITMAP_JPEG,
    FORMAT_BITMAP_PNG,
    FORMAT_BITMAP_WMF,
    FORMAT_BITMAP_EMF,
    FORMAT_BITMAP_DIBINMEMORY,
    FORMAT_BITMAP_TIFF,
    FORMAT_BITMAP_PCX,
    FORMAT_BITMAP_INVALID=0xffffffff
} TYPEFORMATOUT;


/* TResizeFlushPtrFunc : custom realloc buffer function (typically call realloc)
    lpOldPtr : the old pointer buffer
    lpCustomDataForResize : the value gived in FlushDibInBitmap or FlushTif
    dwNewSize :  the size of buffer */
/* example of simple function :
LPVOID AllocResizeFunc(LPVOID lpOldPtr, LPVOID, DWORD dwNewSize)
{
    return realloc(lpOldPtr,dwNewSize);
}
*/
#ifdef __cplusplus
extern "C" {            /* Assume C declarations for C++ */
#endif  /* __cplusplus */
typedef LPVOID (CALLBACK*TResizeFlushPtrFunc)(LPVOID lpOldPtr, 
                                              LPVOID lpCustomDataForResize,
                                              DWORD dwNewSize);


                                      /*
FlushDibInBitmap : Write a DIB in a format of bitmap file, in memory
  lpDib : pointer to Bitmap Header (BITMAPINFO and BITMAPINFOHEADER)
  lpRawData : point to Bitmap raw data (just after Header in lpRawData is NULL)

  lpPtr : pointer where writting data
  dwSize : size of Buffer lpPtr
  
  pResizeFlushPtrFunc : pointer to function to realloc memory of lpPtr buffer 
       is full
  lpCustomDataForResize : custom parameter gived to the pResizeFlushPtrFunc func
  lplpNewPtr : will contain the final pointer to the buffer that contain bitmap file.

  if you are sure the buffer size (in dwSize) is bigger than the bitmap that will be
    created, pResizeFlushPtrFunc, lpCustomDataForResize and lplpNewPtr can be NULL

  if pResizeFlushPtrFunc is not NULL, lpPtr can be NULL and dwSize = 0
  Return value : size of the Bitmap file (or 0 if error)
*/
FLUSHDIBAPIEXPORT DWORD FLUSHDIBAPI FlushDibInBitmap(LPVOID lpDib,LPVOID lpRawData,
                        TYPEFORMATOUT tFormatOut, long lQuality,
                        LPVOID lpPtr,DWORD dwSize,
                        TResizeFlushPtrFunc pResizeFlushPtrFunc=NULL,
                        LPVOID lpCustomDataForResize=NULL,
                        LPVOID* lplpNewPtr=NULL);
#ifdef __cplusplus
}
#endif  /* __cplusplus */


/* class PaintBitmap : class for creating a Bitmap by painting Inside */
/* You create the Object with GetPaintBitmap
   You call Init, with Bitmap Parameters
   After, you can call GetPaintDC, to get the DC.
   You draw in the DC,
   you call FlushTif (see FlushDibInBitmap for understand parameters),
   and you destory the PaintBitmap Object
   */
#ifndef NOPAINTBITMAP

#ifdef __cplusplus
class FLUSHDIBCLASS PAINTBITMAP
{
public:
    PAINTBITMAP() {};
    virtual ~PAINTBITMAP() {};
    virtual BOOL Init(long x,long y,long lNbBitsColor,
                const COLORREF* pcrPalette,long iNbColorPalette,
				long lQuality=0)=0;
    virtual HDC  GetPaintDC()=0;
    virtual DWORD FlushTif(LPVOID lpPtr,DWORD dwSize,
                            TResizeFlushPtrFunc pResizeFlushPtrFunc=NULL,
                            LPVOID lpCustomDataForResize=NULL,
                            LPVOID* lplpNewPtr=NULL)=0;
	virtual const char* GetMimeType() = 0 ;
} ;

typedef PAINTBITMAP* LPPAINTBITMAP;
extern "C" {            /* Assume C declarations for C++ */
#else   /* __cplusplus */
DECLARE_HANDLE(LPPAINTBITMAP)
#endif  /* __cplusplus */

FLUSHDIBAPIEXPORT LPPAINTBITMAP FLUSHDIBAPI GetPaintBitmap(TYPEFORMATOUT tFormatOut);
FLUSHDIBAPIEXPORT BOOL FLUSHDIBAPI InitPaintBitmap(LPPAINTBITMAP, long x,long y,long lNbBitsColor,
                const COLORREF* pcrPalette,long iNbColorPalette,
				long lQuality/*=0*/);
FLUSHDIBAPIEXPORT HDC  FLUSHDIBAPI GetPaintDCPaintBitmap(LPPAINTBITMAP);
FLUSHDIBAPIEXPORT DWORD FLUSHDIBAPI FlushPaintBitmap(LPPAINTBITMAP, LPVOID lpPtr,DWORD dwSize,
                            TResizeFlushPtrFunc pResizeFlushPtrFunc/*=NULL*/,
                            LPVOID lpCustomDataForResize/*=NULL*/,
                            LPVOID* lplpNewPtr/*=NULL*/);
FLUSHDIBAPIEXPORT const char* FLUSHDIBAPI GetMimeTypePaintBitmap(LPPAINTBITMAP) ;
FLUSHDIBAPIEXPORT BOOL FLUSHDIBAPI DeletePaintBitmap(LPPAINTBITMAP) ;
FLUSHDIBAPIEXPORT TYPEFORMATOUT FLUSHDIBAPI GetFormatOutFromExt(LPCSTR szExt);
#ifdef __cplusplus
}
#endif  /* __cplusplus */

#endif // NOPAINTBITMAP
