#include <windows.h>
#include <windowsx.h>
#include "dibtools.h"


/*************************************************************************
 * All the following functions were created by microsoft, they are
 * parts of the sample project "wincap" given with the SDK Win32.
 *
 * Microsoft says that :
 *
 *  You have a royalty-free right to use, modify, reproduce and
 *  distribute the Sample Files (and/or any modified version) in
 *  any way you find useful, provided that you agree that
 *  Microsoft has no warranty obligations or liability for any
 *  Sample Application Files which are modified.
 *
 ************************************************************************/

#define IS_WIN30_DIB(lpbi)  ((*(LPDWORD)(lpbi)) >= sizeof(BITMAPINFOHEADER))
#define CVT(x)      (((x) * 255L) / ((1L<<16)-1))

HANDLE DibTool_CreateDIB(DWORD dwWidth, DWORD dwHeight, DWORD wBitCount)
{
   BITMAPINFOHEADER bi;         // bitmap header
   LPBITMAPINFOHEADER lpbi;     // pointer to BITMAPINFOHEADER
   DWORD dwLen;                 // size of memory block
   HANDLE HANDLE;
   DWORD dwBytesPerLine;        // Number of bytes per scanline


   // Make sure bits per pixel is valid
   if (wBitCount <= 1)
      wBitCount = 1;
   else if (wBitCount <= 4)
      wBitCount = 4;
   else if (wBitCount <= 8)
      wBitCount = 8;
   else if (wBitCount <= 24)
      wBitCount = 24;
   else
      wBitCount = 4;  // set default value to 4 if parameter is bogus

   // initialize BITMAPINFOHEADER
   bi.biSize = sizeof(BITMAPINFOHEADER);
   bi.biWidth = dwWidth;         // fill in width from parameter
   bi.biHeight = dwHeight;       // fill in height from parameter
   bi.biPlanes = 1;              // must be 1
   bi.biBitCount = (WORD)wBitCount;    // from parameter
   bi.biCompression = BI_RGB;    
   bi.biSizeImage = (dwWidth*dwHeight*wBitCount)/8; //0;           // 0's heremean "default"
   bi.biXPelsPerMeter = 2834; //0;
   bi.biYPelsPerMeter = 2834; //0;
   bi.biClrUsed = 0;
   bi.biClrImportant = 0;

   // calculate size of memory block required to store the DIB.  This
   // block should be big enough to hold the BITMAPINFOHEADER, the color
   // table, and the bits

   dwBytesPerLine =   (((wBitCount * dwWidth) + 31) / 32 * 4);
   dwLen = bi.biSize + DibTool_PaletteSize((LPSTR)&bi) + (dwBytesPerLine * dwHeight);

   // alloc memory block to store our bitmap
   HANDLE = GlobalAlloc(GHND, dwLen);

   // major bummer if we couldn't get memory block
   if (!HANDLE)
   {
      return NULL;
   }

   // lock memory and get pointer to it
   lpbi = (LPBITMAPINFOHEADER)GlobalLock(HANDLE);

   // use our bitmap info structure to fill in first part of
   // our DIB with the BITMAPINFOHEADER
   *lpbi = bi;

   // Since we don't know what the colortable and bits should contain,
   // just leave these blank.  Unlock the DIB and return the HANDLE.

   GlobalUnlock(HANDLE);

   /* return handle to the DIB */
   return HANDLE;
}


LPSTR DibTool_FindDIBBits(LPVOID lpDIB)
{
   return (((LPSTR)lpDIB) + (*(LPDWORD)lpDIB) + DibTool_PaletteSize(lpDIB));
}


DWORD DibTool_PaletteSize(LPVOID lpDIB)
{
   /* calculate the size required by the palette */
   if (IS_WIN30_DIB (lpDIB))
      return (DibTool_DIBNumColors(lpDIB) * sizeof(RGBQUAD));
   else
      return (DibTool_DIBNumColors(lpDIB) * sizeof(RGBTRIPLE));
}


DWORD DibTool_DIBNumColors(LPVOID lpDIB)
{
   WORD wBitCount;  // DIB bit count

   /*  If this is a Windows-style DIB, the number of colors in the
    *  color table can be less than the number of bits per pixel
    *  allows for (i.e. lpbi->biClrUsed can be set to some value).
    *  If this is the case, return the appropriate value.
    */

   if (IS_WIN30_DIB(lpDIB))
   {
      DWORD dwClrUsed=0;

      //dwClrUsed = ((LPBITMAPINFOHEADER)lpDIB)->biClrUsed;
      if (dwClrUsed)
     return (WORD)dwClrUsed;
   }

   /*  Calculate the number of colors in the color table based on
    *  the number of bits per pixel for the DIB.
    */
   if (IS_WIN30_DIB(lpDIB))
      wBitCount = ((LPBITMAPINFOHEADER)lpDIB)->biBitCount;
   else
      wBitCount = ((LPBITMAPCOREHEADER)lpDIB)->bcBitCount;

   /* return number of colors based on bits per pixel */
   switch (wBitCount)
      {
   case 1:
      return 2;

   case 4:
      return 16;

   case 8:
      return 256;

   default:
      return 0;
      }
}

/**************************************************************/

// removed : not a dib tools things !
/*
BOOL DibTool_GetFileSize(LPCSTR lpFn,DWORD &dwSize,LPSTR lpTrueName)
{
WIN32_FIND_DATA ffblk;
HANDLE hFind;
char szName[MAX_PATH+1];


  lstrcpy(szName,lpFn);
  dwSize = 0;

  if ((hFind = FindFirstFile(szName,&ffblk)) == INVALID_HANDLE_VALUE)
           return FALSE;
       else
         {
           dwSize = ffblk.nFileSizeLow;
           if (lpTrueName!=NULL)
             lstrcpy(lpTrueName,ffblk.cFileName);
         }
  FindClose(hFind);

  return TRUE;
}
*/

BOOL DIBTool_CompactLine(LPVOID lpDestLine,LPCVOID lpSrcLine,DWORD dwWidth,DWORD dwNbBitPixel)
{
    BOOL fRet= FALSE;
    if (dwNbBitPixel==8)
    {
        if (lpDestLine!=lpSrcLine)
            memcpy(lpDestLine,lpSrcLine,dwWidth);
        fRet=TRUE;
    }

    if (dwNbBitPixel==4)
    {
        unsigned long j;
        //for (j=((dwWidth+1)/2)-1;j>=0;j--)
        for (j=0;j<((dwWidth+1)/2);j++)
        {
            LPBYTE lpSrc = ((LPBYTE)lpSrcLine)+(j*2);
            LPBYTE lpDest = ((LPBYTE)lpDestLine)+(j*1);
            BYTE bValueSrc = *lpSrc;

            *lpDest = (((*(lpSrc+1)) << 0) & 0x0f) |
                      (((*(lpSrc+0)) << 4) & 0xf0);

        }
        fRet=TRUE;
    }

    if (dwNbBitPixel==1)
    {
        unsigned long j;
        //for (j=((dwWidth+7)/8)-1;j>=0;j--)
        for (j=0;j<((dwWidth+7)/8);j++)
        {
            LPBYTE lpSrc = ((LPBYTE)lpSrcLine)+(j*8);
            LPBYTE lpDest = ((LPBYTE)lpDestLine)+(j*1);
            BYTE bValueSrc = *lpSrc;
            *lpDest =  ((*(lpSrc+0) != 0) ? 0x80 : 0) |
                       ((*(lpSrc+1) != 0) ? 0x40 : 0) | 
                       ((*(lpSrc+2) != 0) ? 0x20 : 0) | 
                       ((*(lpSrc+3) != 0) ? 0x10 : 0) | 
                       ((*(lpSrc+4) != 0) ? 0x08 : 0) | 
                       ((*(lpSrc+5) != 0) ? 0x04 : 0) | 
                       ((*(lpSrc+6) != 0) ? 0x02 : 0) | 
                       ((*(lpSrc+7) != 0) ? 0x01 : 0) ;

        }
        fRet=TRUE;
    }
    return fRet;
}

BOOL DIBTool_GetDIBInfo(LPVOID lpDib,
                        long &lX,long &lY,long &lNbBitsColor,
	                    long &iNbColorPalette,
                        LPBYTE & rawdata)
{
    BOOL fRet = FALSE;
    LPBITMAPINFO lpBmi=(LPBITMAPINFO)lpDib;
    if (lpBmi != NULL)
    {
        lX = lpBmi -> bmiHeader.biWidth;
        lY = lpBmi -> bmiHeader.biHeight ;
        lNbBitsColor = lpBmi -> bmiHeader.biBitCount;
        iNbColorPalette = DibTool_DIBNumColors(lpDib);
        rawdata = (LPBYTE)DibTool_FindDIBBits(lpDib) ;
        fRet = TRUE;
    }
    return fRet;
}

BOOL DIBTool_AnalysePalette(RGBQUAD* lpPalette, DWORD dwNbBits,
                               BOOL &fIsPalette, DWORD &dwNbPalEntry,
                               BOOL &fIsGrayPalette, BOOL &fIsOrderedGrayPalette,
                               DWORD& dwStepPalInGray)
{
    dwNbPalEntry=dwStepPalInGray=0;
    fIsPalette=fIsGrayPalette=fIsOrderedGrayPalette=FALSE;
    BOOL fRet = fIsPalette = (dwNbBits<16);
    if (fRet)
        dwNbPalEntry = 1<<dwNbBits;

    if (dwNbPalEntry>0)
    {
        DWORD i;
        fIsGrayPalette = fIsOrderedGrayPalette=TRUE;
        dwStepPalInGray = 0x100 / dwNbPalEntry;
        for (i=0;(i<dwNbPalEntry) && fIsGrayPalette;i++,lpPalette++)
        {
            BYTE bCurEntryValue = lpPalette->rgbBlue;
            fIsGrayPalette = (bCurEntryValue == lpPalette->rgbGreen) &&
                             (bCurEntryValue == lpPalette->rgbRed);
            if ((fIsOrderedGrayPalette) && (fIsGrayPalette))
                fIsOrderedGrayPalette = (bCurEntryValue== ((BYTE)((i*255)/(dwNbPalEntry-1))));
        }
        if (!fIsGrayPalette)
            fIsOrderedGrayPalette=FALSE;
    }
    return (fRet);
}


BOOL DIBTool_ExpandLine(LPVOID lpDestLine,LPCVOID lpSrcLine,DWORD dwWidth,DWORD dwNbBitPixel)
{
    BOOL fRet= FALSE;
    if (dwNbBitPixel==8)
    {
        if (lpDestLine!=lpSrcLine)
            memcpy(lpDestLine,lpSrcLine,dwWidth);
        fRet=TRUE;
    }

    if (dwNbBitPixel==4)
    {
        signed long j;
        for (j=((dwWidth+1)/2)-1;j>=0;j--)
        {
            LPBYTE lpSrc = ((LPBYTE)lpSrcLine)+j;
            LPBYTE lpDest = ((LPBYTE)lpDestLine)+(j*2);
            BYTE bValueSrc = *lpSrc;
            *(lpDest+0) = bValueSrc / 0x10;
            *(lpDest+1) = bValueSrc & 0x0f;
        }
        fRet=TRUE;
    }

    if (dwNbBitPixel==1)
    {
        signed long j;
        for (j=((dwWidth+7)/8)-1;j>=0;j--)
        {
            LPBYTE lpSrc = ((LPBYTE)lpSrcLine)+j;
            LPBYTE lpDest = ((LPBYTE)lpDestLine)+(j*8);
            BYTE bValueSrc = *lpSrc;
            *(lpDest+0) = ((bValueSrc & 0x80) == 0) ? 0 : 1;
            *(lpDest+1) = ((bValueSrc & 0x40) == 0) ? 0 : 1;
            *(lpDest+2) = ((bValueSrc & 0x20) == 0) ? 0 : 1;
            *(lpDest+3) = ((bValueSrc & 0x10) == 0) ? 0 : 1;
            *(lpDest+4) = ((bValueSrc & 0x08) == 0) ? 0 : 1;
            *(lpDest+5) = ((bValueSrc & 0x04) == 0) ? 0 : 1;
            *(lpDest+6) = ((bValueSrc & 0x02) == 0) ? 0 : 1;
            *(lpDest+7) = ((bValueSrc & 0x01) == 0) ? 0 : 1;
        }
        fRet=TRUE;
    }
    return fRet;
}
